from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from cassandra import DriverException, OperationTimedOut
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import (
    EXEC_PROFILE_DEFAULT,
    Cluster,
    ExecutionProfile,
    ProtocolVersion,
    Session,
)

from datahub.ingestion.api.source import SourceReport
from datahub.ingestion.source.cassandra.cassandra_config import CassandraSourceConfig


@dataclass
class CassandraKeyspace:
    keyspace_name: str
    durable_writes: bool
    replication: Dict


@dataclass
class CassandraTable:
    keyspace_name: str
    table_name: str
    bloom_filter_fp_chance: Optional[float]
    caching: Optional[Dict[str, str]]
    comment: Optional[str]
    compaction: Optional[Dict[str, Any]]
    compression: Optional[Dict[str, Any]]
    crc_check_chance: Optional[float]
    dclocal_read_repair_chance: Optional[float]
    default_time_to_live: Optional[int]
    extensions: Optional[Dict[str, Any]]
    gc_grace_seconds: Optional[int]
    max_index_interval: Optional[int]
    memtable_flush_period_in_ms: Optional[int]
    min_index_interval: Optional[int]
    read_repair_chance: Optional[float]
    speculative_retry: Optional[str]


@dataclass
class CassandraColumn:
    keyspace_name: str
    table_name: str
    column_name: str
    type: str
    clustering_order: Optional[str]
    kind: Optional[str]
    position: Optional[int]


@dataclass
class CassandraView(CassandraTable):
    view_name: str
    include_all_columns: Optional[bool]
    where_clause: str = ""


@dataclass
class CassandraEntities:
    keyspaces: List[str] = field(default_factory=list)
    tables: Dict[str, List[str]] = field(
        default_factory=dict
    )  # Maps keyspace -> tables
    columns: Dict[str, List[CassandraColumn]] = field(
        default_factory=dict
    )  # Maps tables -> columns


# - Referencing system_schema: https://docs.datastax.com/en/cql-oss/3.x/cql/cql_using/useQuerySystem.html#Table3.ColumnsinSystem_SchemaTables-Cassandra3.0 - #
# this keyspace contains details about the cassandra cluster's keyspaces, tables, and columns


class CassandraQueries:
    # get all keyspaces
    GET_KEYSPACES_QUERY = "SELECT * FROM system_schema.keyspaces"
    # get all tables for a keyspace
    GET_TABLES_QUERY = "SELECT * FROM system_schema.tables WHERE keyspace_name = %s"
    # get all columns for a table
    GET_COLUMNS_QUERY = "SELECT * FROM system_schema.columns WHERE keyspace_name = %s AND table_name = %s"
    # get all views for a keyspace
    GET_VIEWS_QUERY = "SELECT * FROM system_schema.views WHERE keyspace_name = %s"
    # Row Count
    ROW_COUNT = 'SELECT COUNT(*) AS row_count FROM {}."{}"'
    # Column Count
    COLUMN_COUNT = "SELECT COUNT(*) AS column_count FROM system_schema.columns WHERE keyspace_name = '{}' AND table_name = '{}'"


class CassandraAPI:
    def __init__(self, config: CassandraSourceConfig, report: SourceReport):
        self.config = config
        self.report = report
        self._cassandra_session: Optional[Session] = None

    def authenticate(self) -> bool:
        """Establish a connection to Cassandra and return the session."""
        try:
            if self.config.cloud_config:
                cloud_config = self.config.cloud_config
                cluster_cloud_config = {
                    "connect_timeout": cloud_config.connect_timeout,
                    "use_default_tempdir": True,
                    "secure_connect_bundle": cloud_config.secure_connect_bundle,
                }
                profile = ExecutionProfile(request_timeout=cloud_config.request_timeout)
                auth_provider = PlainTextAuthProvider(
                    "token",
                    cloud_config.token,
                )
                cluster = Cluster(
                    cloud=cluster_cloud_config,
                    auth_provider=auth_provider,
                    execution_profiles={EXEC_PROFILE_DEFAULT: profile},
                    protocol_version=ProtocolVersion.V4,
                )

                self._cassandra_session = cluster.connect()
                return True
            if self.config.username and self.config.password:
                auth_provider = PlainTextAuthProvider(
                    username=self.config.username, password=self.config.password
                )
                cluster = Cluster(
                    [self.config.contact_point],
                    port=self.config.port,
                    auth_provider=auth_provider,
                    load_balancing_policy=None,
                )
            else:
                cluster = Cluster(
                    [self.config.contact_point],
                    port=self.config.port,
                    load_balancing_policy=None,
                )

            self._cassandra_session = cluster.connect()
            return True
        except OperationTimedOut as e:
            self.report.failure(
                message="Failed to Authenticate", context=f"{str(e.errors)}", exc=e
            )
            return False
        except DriverException as e:
            self.report.failure(message="Failed to Authenticate", exc=e)
            return False
        except Exception as e:
            self.report.failure(message="Failed to authenticate to Cassandra", exc=e)
            return False

    def get(self, query: str, parameters: Optional[List] = []) -> List:
        if not self._cassandra_session:
            return []

        resp = self._cassandra_session.execute(query, parameters)
        return resp

    def get_keyspaces(self) -> List[CassandraKeyspace]:
        """Fetch all keyspaces."""
        try:
            keyspaces = self.get(CassandraQueries.GET_KEYSPACES_QUERY)
            keyspace_list = [
                CassandraKeyspace(
                    keyspace_name=row.keyspace_name,
                    durable_writes=row.durable_writes,
                    replication=dict(row.replication),
                )
                for row in keyspaces
            ]
            return keyspace_list
        except DriverException as e:
            self.report.warning(
                message="Failed to fetch keyspaces", context=f"{str(e)}", exc=e
            )
            return []
        except Exception as e:
            self.report.warning(message="Failed to fetch keyspaces", exc=e)
            return []

    def get_tables(self, keyspace_name: str) -> List[CassandraTable]:
        """Fetch all tables for a given keyspace."""
        try:
            tables = self.get(CassandraQueries.GET_TABLES_QUERY, [keyspace_name])
            table_list = [
                CassandraTable(
                    keyspace_name=row.keyspace_name,
                    table_name=row.table_name,
                    bloom_filter_fp_chance=row.bloom_filter_fp_chance,
                    caching=dict(row.caching),
                    comment=row.comment,
                    compaction=dict(row.compaction),
                    compression=dict(row.compression),
                    crc_check_chance=row.crc_check_chance,
                    dclocal_read_repair_chance=row.dclocal_read_repair_chance,
                    default_time_to_live=row.default_time_to_live,
                    extensions=dict(row.extensions),
                    gc_grace_seconds=row.gc_grace_seconds,
                    max_index_interval=row.max_index_interval,
                    memtable_flush_period_in_ms=row.memtable_flush_period_in_ms,
                    min_index_interval=row.min_index_interval,
                    read_repair_chance=row.read_repair_chance,
                    speculative_retry=row.speculative_retry,
                )
                for row in tables
            ]
            return table_list
        except DriverException as e:
            self.report.warning(
                message="Failed to fetch tables for keyspace",
                context=f"{str(e)}",
                exc=e,
            )
            return []
        except Exception as e:
            self.report.warning(
                message="Failed to fetch tables for keyspace",
                context=f"{keyspace_name}",
                exc=e,
            )
            return []

    def get_columns(self, keyspace_name: str, table_name: str) -> List[CassandraColumn]:
        """Fetch all columns for a given table."""
        try:
            column_infos = self.get(
                CassandraQueries.GET_COLUMNS_QUERY, [keyspace_name, table_name]
            )
            column_list = [
                CassandraColumn(
                    keyspace_name=row.keyspace_name,
                    table_name=row.table_name,
                    column_name=row.column_name,
                    clustering_order=row.clustering_order,
                    kind=row.kind,
                    position=row.position,
                    type=row.type,
                )
                for row in column_infos
            ]
            return column_list
        except DriverException as e:
            self.report.warning(
                message="Failed to fetch columns for table", context=f"{str(e)}", exc=e
            )
            return []
        except Exception as e:
            self.report.warning(
                message="Failed to fetch columns for table",
                context=f"{keyspace_name}.{table_name}",
                exc=e,
            )
            return []

    def get_views(self, keyspace_name: str) -> List[CassandraView]:
        """Fetch all views for a given keyspace."""
        try:
            views = self.get(CassandraQueries.GET_VIEWS_QUERY, [keyspace_name])
            view_list = [
                CassandraView(
                    table_name=row.base_table_name,
                    keyspace_name=row.keyspace_name,
                    view_name=row.view_name,
                    bloom_filter_fp_chance=row.bloom_filter_fp_chance,
                    caching=dict(row.caching),
                    comment=row.comment,
                    compaction=dict(row.compaction),
                    compression=dict(row.compression),
                    crc_check_chance=row.crc_check_chance,
                    dclocal_read_repair_chance=row.dclocal_read_repair_chance,
                    default_time_to_live=row.default_time_to_live,
                    extensions=dict(row.extensions),
                    gc_grace_seconds=row.gc_grace_seconds,
                    include_all_columns=row.include_all_columns,
                    max_index_interval=row.max_index_interval,
                    memtable_flush_period_in_ms=row.memtable_flush_period_in_ms,
                    min_index_interval=row.min_index_interval,
                    read_repair_chance=row.read_repair_chance,
                    speculative_retry=row.speculative_retry,
                    where_clause=row.where_clause,
                )
                for row in views
            ]
            return view_list
        except DriverException as e:
            self.report.warning(
                message="Failed to fetch views for keyspace", context=f"{str(e)}", exc=e
            )
            return []
        except Exception as e:
            self.report.warning(
                message="Failed to fetch views for keyspace",
                context=f"{keyspace_name}",
                exc=e,
            )
            return []

    def execute(self, query: str, limit: Optional[int] = None) -> List:
        """Fetch stats for cassandra"""
        try:
            if not self._cassandra_session:
                return []
            if limit:
                query = query + f" LIMIT {limit}"
            result_set = self._cassandra_session.execute(query).all()
            return result_set
        except DriverException as e:
            self.report.warning(
                message="Failed to fetch stats for keyspace", context=str(e), exc=e
            )
            return []
        except Exception:
            self.report.warning(
                message="Failed to fetch stats for keyspace",
                context=f"{query}",
            )
            return []

    def close(self):
        """Close the Cassandra session."""
        if self._cassandra_session:
            self._cassandra_session.shutdown()



import dataclasses
import json
import logging
from typing import Any, Dict, Iterable, List, Optional

from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
    make_schema_field_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import (
    ContainerKey,
    add_dataset_to_container,
    gen_containers,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.cassandra.cassandra_api import (
    CassandraAPI,
    CassandraColumn,
    CassandraEntities,
    CassandraKeyspace,
    CassandraTable,
    CassandraView,
)
from datahub.ingestion.source.cassandra.cassandra_config import CassandraSourceConfig
from datahub.ingestion.source.cassandra.cassandra_profiling import CassandraProfiler
from datahub.ingestion.source.cassandra.cassandra_utils import (
    SYSTEM_KEYSPACE_LIST,
    CassandraSourceReport,
    CassandraToSchemaFieldConverter,
)
from datahub.ingestion.source.common.subtypes import (
    DatasetContainerSubTypes,
    DatasetSubTypes,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import StatusClass
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    SchemaField,
    SchemaMetadata,
)
from datahub.metadata.schema_classes import (
    DataPlatformInstanceClass,
    DatasetLineageTypeClass,
    DatasetPropertiesClass,
    FineGrainedLineageClass,
    FineGrainedLineageDownstreamTypeClass,
    FineGrainedLineageUpstreamTypeClass,
    OtherSchemaClass,
    SubTypesClass,
    UpstreamClass,
    UpstreamLineageClass,
    ViewPropertiesClass,
)

logger = logging.getLogger(__name__)

PLATFORM_NAME_IN_DATAHUB = "cassandra"


class KeyspaceKey(ContainerKey):
    keyspace: str


@platform_name("Cassandra")
@config_class(CassandraSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.CONTAINERS, "Enabled by default")
@capability(SourceCapability.SCHEMA_METADATA, "Enabled by default")
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(
    SourceCapability.DELETION_DETECTION,
    "Optionally enabled via `stateful_ingestion.remove_stale_metadata`",
    supported=True,
)
class CassandraSource(StatefulIngestionSourceBase):
    """
    This plugin extracts the following:

    - Metadata for tables
    - Column types associated with each table column
    - The keyspace each table belongs to
    """

    config: CassandraSourceConfig
    report: CassandraSourceReport
    platform: str

    def __init__(self, ctx: PipelineContext, config: CassandraSourceConfig):
        super().__init__(config, ctx)
        self.ctx = ctx
        self.platform = PLATFORM_NAME_IN_DATAHUB
        self.config = config
        self.report = CassandraSourceReport()
        self.cassandra_api = CassandraAPI(config, self.report)
        self.cassandra_data = CassandraEntities()
        # For profiling
        self.profiler = CassandraProfiler(config, self.report, self.cassandra_api)

    @classmethod
    def create(cls, config_dict, ctx):
        config = CassandraSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_platform(self) -> str:
        return PLATFORM_NAME_IN_DATAHUB

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        if not self.cassandra_api.authenticate():
            return
        keyspaces: List[CassandraKeyspace] = self.cassandra_api.get_keyspaces()
        for keyspace in keyspaces:
            keyspace_name: str = keyspace.keyspace_name
            if keyspace_name in SYSTEM_KEYSPACE_LIST:
                continue

            if not self.config.keyspace_pattern.allowed(keyspace_name):
                self.report.report_dropped(keyspace_name)
                continue

            yield from self._generate_keyspace_container(keyspace)

            try:
                yield from self._extract_tables_from_keyspace(keyspace_name)
            except Exception as e:
                self.report.num_tables_failed += 1
                self.report.failure(
                    message="Failed to extract table metadata for keyspace",
                    context=keyspace_name,
                    exc=e,
                )
            try:
                yield from self._extract_views_from_keyspace(keyspace_name)
            except Exception as e:
                self.report.num_views_failed += 1
                self.report.failure(
                    message="Failed to extract view metadata for keyspace ",
                    context=keyspace_name,
                    exc=e,
                )

        # Profiling
        if self.config.is_profiling_enabled():
            yield from self.profiler.get_workunits(self.cassandra_data)

    def _generate_keyspace_container(
        self, keyspace: CassandraKeyspace
    ) -> Iterable[MetadataWorkUnit]:
        keyspace_container_key = self._generate_keyspace_container_key(
            keyspace.keyspace_name
        )
        yield from gen_containers(
            container_key=keyspace_container_key,
            name=keyspace.keyspace_name,
            qualified_name=keyspace.keyspace_name,
            extra_properties={
                "durable_writes": str(keyspace.durable_writes),
                "replication": json.dumps(keyspace.replication),
            },
            sub_types=[DatasetContainerSubTypes.KEYSPACE],
        )

    def _generate_keyspace_container_key(self, keyspace_name: str) -> ContainerKey:
        return KeyspaceKey(
            keyspace=keyspace_name,
            platform=self.platform,
            instance=self.config.platform_instance,
            env=self.config.env,
        )

    # get all tables for a given keyspace, iterate over them to extract column metadata
    def _extract_tables_from_keyspace(
        self, keyspace_name: str
    ) -> Iterable[MetadataWorkUnit]:
        self.cassandra_data.keyspaces.append(keyspace_name)
        tables: List[CassandraTable] = self.cassandra_api.get_tables(keyspace_name)
        for table in tables:
            # define the dataset urn for this table to be used downstream
            table_name: str = table.table_name
            dataset_name: str = f"{keyspace_name}.{table_name}"

            if not self.config.table_pattern.allowed(dataset_name):
                self.report.report_dropped(dataset_name)
                continue

            self.cassandra_data.tables.setdefault(keyspace_name, []).append(table_name)
            self.report.report_entity_scanned(dataset_name, ent_type="Table")

            dataset_urn = make_dataset_urn_with_platform_instance(
                platform=self.platform,
                name=dataset_name,
                env=self.config.env,
                platform_instance=self.config.platform_instance,
            )

            # 1. Extract columns from table, then construct and emit the schemaMetadata aspect.
            try:
                yield from self._extract_columns_from_table(
                    keyspace_name, table_name, dataset_urn
                )
            except Exception as e:
                self.report.failure(
                    message="Failed to extract columns from table",
                    context=table_name,
                    exc=e,
                )

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=StatusClass(removed=False),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=SubTypesClass(
                    typeNames=[
                        DatasetSubTypes.TABLE,
                    ]
                ),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=DatasetPropertiesClass(
                    name=table_name,
                    qualifiedName=f"{keyspace_name}.{table_name}",
                    description=table.comment,
                    customProperties={
                        "bloom_filter_fp_chance": str(table.bloom_filter_fp_chance),
                        "caching": json.dumps(table.caching),
                        "compaction": json.dumps(table.compaction),
                        "compression": json.dumps(table.compression),
                        "crc_check_chance": str(table.crc_check_chance),
                        "dclocal_read_repair_chance": str(
                            table.dclocal_read_repair_chance
                        ),
                        "default_time_to_live": str(table.default_time_to_live),
                        "extensions": json.dumps(table.extensions),
                        "gc_grace_seconds": str(table.gc_grace_seconds),
                        "max_index_interval": str(table.max_index_interval),
                        "min_index_interval": str(table.min_index_interval),
                        "memtable_flush_period_in_ms": str(
                            table.memtable_flush_period_in_ms
                        ),
                        "read_repair_chance": str(table.read_repair_chance),
                        "speculative_retry": str(table.speculative_retry),
                    },
                ),
            ).as_workunit()

            yield from add_dataset_to_container(
                container_key=self._generate_keyspace_container_key(keyspace_name),
                dataset_urn=dataset_urn,
            )

            if self.config.platform_instance:
                yield MetadataChangeProposalWrapper(
                    entityUrn=dataset_urn,
                    aspect=DataPlatformInstanceClass(
                        platform=make_data_platform_urn(self.platform),
                        instance=make_dataplatform_instance_urn(
                            self.platform, self.config.platform_instance
                        ),
                    ),
                ).as_workunit()

    # get all columns for a given table, iterate over them to extract column metadata
    def _extract_columns_from_table(
        self, keyspace_name: str, table_name: str, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        column_infos: List[CassandraColumn] = self.cassandra_api.get_columns(
            keyspace_name, table_name
        )
        schema_fields: List[SchemaField] = list(
            CassandraToSchemaFieldConverter.get_schema_fields(column_infos)
        )
        if not schema_fields:
            self.report.report_warning(
                message="Table has no columns, skipping", context=table_name
            )
            return

        jsonable_column_infos: List[Dict[str, Any]] = []
        for column in column_infos:
            self.cassandra_data.columns.setdefault(table_name, []).append(column)
            jsonable_column_infos.append(dataclasses.asdict(column))

        schema_metadata: SchemaMetadata = SchemaMetadata(
            schemaName=table_name,
            platform=make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=OtherSchemaClass(
                rawSchema=json.dumps(jsonable_column_infos)
            ),
            fields=schema_fields,
        )

        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=schema_metadata,
        ).as_workunit()

    def _extract_views_from_keyspace(
        self, keyspace_name: str
    ) -> Iterable[MetadataWorkUnit]:
        views: List[CassandraView] = self.cassandra_api.get_views(keyspace_name)
        for view in views:
            view_name: str = view.view_name
            dataset_name: str = f"{keyspace_name}.{view_name}"
            self.report.report_entity_scanned(dataset_name)
            dataset_urn: str = make_dataset_urn_with_platform_instance(
                platform=self.platform,
                name=dataset_name,
                env=self.config.env,
                platform_instance=self.config.platform_instance,
            )

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=StatusClass(removed=False),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=SubTypesClass(
                    typeNames=[
                        DatasetSubTypes.VIEW,
                    ]
                ),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=ViewPropertiesClass(
                    materialized=True,
                    viewLogic=view.where_clause,  # Use the WHERE clause as view logic
                    viewLanguage="CQL",  # Use "CQL" as the language
                ),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=DatasetPropertiesClass(
                    name=view_name,
                    qualifiedName=f"{keyspace_name}.{view_name}",
                    description=view.comment,
                    customProperties={
                        "bloom_filter_fp_chance": str(view.bloom_filter_fp_chance),
                        "caching": json.dumps(view.caching),
                        "compaction": json.dumps(view.compaction),
                        "compression": json.dumps(view.compression),
                        "crc_check_chance": str(view.crc_check_chance),
                        "include_all_columns": str(view.include_all_columns),
                        "dclocal_read_repair_chance": str(
                            view.dclocal_read_repair_chance
                        ),
                        "default_time_to_live": str(view.default_time_to_live),
                        "extensions": json.dumps(view.extensions),
                        "gc_grace_seconds": str(view.gc_grace_seconds),
                        "max_index_interval": str(view.max_index_interval),
                        "min_index_interval": str(view.min_index_interval),
                        "memtable_flush_period_in_ms": str(
                            view.memtable_flush_period_in_ms
                        ),
                        "read_repair_chance": str(view.read_repair_chance),
                        "speculative_retry": str(view.speculative_retry),
                    },
                ),
            ).as_workunit()

            try:
                yield from self._extract_columns_from_table(
                    keyspace_name, view_name, dataset_urn
                )
            except Exception as e:
                self.report.failure(
                    message="Failed to extract columns from views",
                    context=view_name,
                    exc=e,
                )

            # Construct and emit lineage off of 'base_table_name'
            # NOTE: we don't need to use 'base_table_id' since table is always in same keyspace, see https://docs.datastax.com/en/cql-oss/3.3/cql/cql_reference/cqlCreateMaterializedView.html#cqlCreateMaterializedView__keyspace-name
            upstream_urn: str = make_dataset_urn_with_platform_instance(
                platform=self.platform,
                name=f"{keyspace_name}.{view.table_name}",
                env=self.config.env,
                platform_instance=self.config.platform_instance,
            )
            fineGrainedLineages = self.get_upstream_fields_of_field_in_datasource(
                view_name, dataset_urn, upstream_urn
            )
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=UpstreamLineageClass(
                    upstreams=[
                        UpstreamClass(
                            dataset=upstream_urn,
                            type=DatasetLineageTypeClass.VIEW,
                        )
                    ],
                    fineGrainedLineages=fineGrainedLineages,
                ),
            ).as_workunit()

            yield from add_dataset_to_container(
                container_key=self._generate_keyspace_container_key(keyspace_name),
                dataset_urn=dataset_urn,
            )

            if self.config.platform_instance:
                yield MetadataChangeProposalWrapper(
                    entityUrn=dataset_urn,
                    aspect=DataPlatformInstanceClass(
                        platform=make_data_platform_urn(self.platform),
                        instance=make_dataplatform_instance_urn(
                            self.platform, self.config.platform_instance
                        ),
                    ),
                ).as_workunit()

    def get_upstream_fields_of_field_in_datasource(
        self, table_name: str, dataset_urn: str, upstream_urn: str
    ) -> List[FineGrainedLineageClass]:
        column_infos = self.cassandra_data.columns.get(table_name, [])
        # Collect column-level lineage
        fine_grained_lineages = []
        for column_info in column_infos:
            source_column = column_info.column_name
            if source_column:
                fine_grained_lineages.append(
                    FineGrainedLineageClass(
                        upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                        downstreamType=FineGrainedLineageDownstreamTypeClass.FIELD,
                        downstreams=[make_schema_field_urn(dataset_urn, source_column)],
                        upstreams=[make_schema_field_urn(upstream_urn, source_column)],
                    )
                )
        return fine_grained_lineages

    def get_report(self):
        return self.report

    def close(self):
        self.cassandra_api.close()
        super().close()



import copy
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Type, cast

from datahub.configuration.common import (
    TransformerSemantics,
    TransformerSemanticsConfigModel,
)
from datahub.configuration.import_resolver import pydantic_resolve_key
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetPropertiesTransformer,
)
from datahub.metadata.schema_classes import DatasetPropertiesClass


class AddDatasetPropertiesResolverBase(ABC):
    @abstractmethod
    def get_properties_to_add(self, entity_urn: str) -> Dict[str, str]:
        pass


class AddDatasetPropertiesConfig(TransformerSemanticsConfigModel):
    add_properties_resolver_class: Type[AddDatasetPropertiesResolverBase]

    class Config:
        arbitrary_types_allowed = True

    _resolve_properties_class = pydantic_resolve_key("add_properties_resolver_class")


class AddDatasetProperties(DatasetPropertiesTransformer):
    """Transformer that adds properties to datasets according to a callback function."""

    ctx: PipelineContext
    config: AddDatasetPropertiesConfig

    def __init__(
        self,
        config: AddDatasetPropertiesConfig,
        ctx: PipelineContext,
        **resolver_args: Dict[str, Any],
    ):
        super().__init__()
        self.ctx = ctx
        self.config = config
        self.resolver_args = resolver_args

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "AddDatasetProperties":
        config = AddDatasetPropertiesConfig.parse_obj(config_dict)
        return cls(config, ctx)

    @staticmethod
    def _merge_with_server_properties(
        graph: DataHubGraph,
        entity_urn: str,
        dataset_properties_aspect: Optional[DatasetPropertiesClass],
    ) -> Optional[DatasetPropertiesClass]:
        assert dataset_properties_aspect

        server_dataset_properties_aspect: Optional[DatasetPropertiesClass] = (
            graph.get_dataset_properties(entity_urn)
        )
        # No need to take any action if server properties is None or there is not customProperties in server properties
        if (
            server_dataset_properties_aspect is None
            or not server_dataset_properties_aspect.customProperties
        ):
            return dataset_properties_aspect

        custom_properties_to_add = server_dataset_properties_aspect.customProperties
        # Give precedence to ingestion custom properties
        # if same property exist on server as well as in input aspect then value of input aspect would get set in output
        custom_properties_to_add.update(dataset_properties_aspect.customProperties)

        patch_dataset_properties: DatasetPropertiesClass = copy.deepcopy(
            dataset_properties_aspect
        )
        patch_dataset_properties.customProperties = custom_properties_to_add

        return patch_dataset_properties

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        in_dataset_properties_aspect: DatasetPropertiesClass = cast(
            DatasetPropertiesClass, aspect
        )
        if not in_dataset_properties_aspect:
            in_dataset_properties_aspect = DatasetPropertiesClass()

        out_dataset_properties_aspect: DatasetPropertiesClass = copy.deepcopy(
            in_dataset_properties_aspect
        )
        if self.config.replace_existing is True:
            # clean the existing properties
            out_dataset_properties_aspect.customProperties = {}

        properties_to_add = self.config.add_properties_resolver_class(  # type: ignore
            **self.resolver_args
        ).get_properties_to_add(entity_urn)

        out_dataset_properties_aspect.customProperties.update(properties_to_add)
        if self.config.semantics == TransformerSemantics.PATCH:
            assert self.ctx.graph
            patch_dataset_properties_aspect = (
                AddDatasetProperties._merge_with_server_properties(
                    self.ctx.graph, entity_urn, out_dataset_properties_aspect
                )
            )
            return cast(Optional[Aspect], patch_dataset_properties_aspect)

        return cast(Aspect, out_dataset_properties_aspect)


class SimpleAddDatasetPropertiesConfig(TransformerSemanticsConfigModel):
    properties: Dict[str, str]


class SimpleAddDatasetPropertiesResolverClass(AddDatasetPropertiesResolverBase):
    def __init__(self, properties: Dict[str, str]):
        self.properties = properties

    def get_properties_to_add(self, entity_urn: str) -> Dict[str, str]:
        return self.properties


class SimpleAddDatasetProperties(AddDatasetProperties):
    """Transformer that adds a specified set of properties to each dataset."""

    def __init__(self, config: SimpleAddDatasetPropertiesConfig, ctx: PipelineContext):
        generic_config = AddDatasetPropertiesConfig(
            add_properties_resolver_class=SimpleAddDatasetPropertiesResolverClass,
            replace_existing=config.replace_existing,
            semantics=config.semantics,
        )
        resolver_args = {"properties": config.properties}
        super().__init__(generic_config, ctx, **resolver_args)

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "SimpleAddDatasetProperties":
        config = SimpleAddDatasetPropertiesConfig.parse_obj(config_dict)
        return cls(config, ctx)



import warnings
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import TypeAlias

import datahub.metadata.schema_classes as models
from datahub.emitter.mce_builder import (
    make_ts_millis,
    make_user_urn,
    parse_ts_millis,
    validate_ownership_type,
)
from datahub.emitter.mcp_builder import ContainerKey
from datahub.errors import MultipleSubtypesWarning, SdkUsageError
from datahub.metadata.urns import (
    CorpGroupUrn,
    CorpUserUrn,
    DataJobUrn,
    DataPlatformInstanceUrn,
    DataPlatformUrn,
    DatasetUrn,
    DomainUrn,
    GlossaryTermUrn,
    OwnershipTypeUrn,
    TagUrn,
    Urn,
)
from datahub.sdk._entity import Entity
from datahub.utilities.urns.error import InvalidUrnError

if TYPE_CHECKING:
    from datahub.sdk.container import Container

UrnOrStr: TypeAlias = Union[Urn, str]
DatasetUrnOrStr: TypeAlias = Union[str, DatasetUrn]
DatajobUrnOrStr: TypeAlias = Union[str, DataJobUrn]

ActorUrn: TypeAlias = Union[CorpUserUrn, CorpGroupUrn]


def make_time_stamp(ts: Optional[datetime]) -> Optional[models.TimeStampClass]:
    if ts is None:
        return None
    return models.TimeStampClass(time=make_ts_millis(ts))


def parse_time_stamp(ts: Optional[models.TimeStampClass]) -> Optional[datetime]:
    if ts is None:
        return None
    return parse_ts_millis(ts.time)


class HasPlatformInstance(Entity):
    __slots__ = ()

    def _set_platform_instance(
        self,
        platform: Union[str, DataPlatformUrn],
        instance: Union[None, str, DataPlatformInstanceUrn],
    ) -> None:
        platform = DataPlatformUrn(platform)
        if instance is not None:
            try:
                instance = DataPlatformInstanceUrn.from_string(instance)
            except InvalidUrnError:
                if not isinstance(
                    instance, DataPlatformInstanceUrn
                ):  # redundant check to make mypy happy
                    instance = DataPlatformInstanceUrn(platform, instance)
        # At this point, instance is either None or a DataPlatformInstanceUrn.

        self._set_aspect(
            models.DataPlatformInstanceClass(
                platform=platform.urn(),
                instance=instance.urn() if instance else None,
            )
        )

    @property
    def platform_instance(self) -> Optional[DataPlatformInstanceUrn]:
        dataPlatformInstance = self._get_aspect(models.DataPlatformInstanceClass)
        if dataPlatformInstance and dataPlatformInstance.instance:
            return DataPlatformInstanceUrn.from_string(dataPlatformInstance.instance)
        return None


class HasSubtype(Entity):
    __slots__ = ()

    @property
    def subtype(self) -> Optional[str]:
        subtypes = self._get_aspect(models.SubTypesClass)
        if subtypes and subtypes.typeNames:
            if len(subtypes.typeNames) > 1:
                warnings.warn(
                    f"The entity {self.urn} has multiple subtypes: {subtypes.typeNames}. "
                    "Only the first subtype will be considered.",
                    MultipleSubtypesWarning,
                    stacklevel=2,
                )
            return subtypes.typeNames[0]
        return None

    def set_subtype(self, subtype: str) -> None:
        self._set_aspect(models.SubTypesClass(typeNames=[subtype]))


OwnershipTypeType: TypeAlias = Union[str, OwnershipTypeUrn]
OwnerInputType: TypeAlias = Union[
    str,
    ActorUrn,
    Tuple[Union[str, ActorUrn], OwnershipTypeType],
    models.OwnerClass,
]
OwnersInputType: TypeAlias = List[OwnerInputType]


class HasOwnership(Entity):
    __slots__ = ()

    @staticmethod
    def _parse_owner_class(owner: OwnerInputType) -> models.OwnerClass:
        if isinstance(owner, models.OwnerClass):
            return owner

        owner_type = models.OwnershipTypeClass.TECHNICAL_OWNER
        owner_type_urn = None

        if isinstance(owner, tuple):
            raw_owner, raw_owner_type = owner

            if isinstance(raw_owner_type, OwnershipTypeUrn):
                owner_type = models.OwnershipTypeClass.CUSTOM
                owner_type_urn = str(raw_owner_type)
            else:
                owner_type, owner_type_urn = validate_ownership_type(raw_owner_type)
        else:
            raw_owner = owner

        if isinstance(raw_owner, str):
            # Tricky: this will gracefully handle a user passing in a group urn as a string.
            # TODO: is this the right behavior? or should we require a valid urn here?
            return models.OwnerClass(
                owner=make_user_urn(raw_owner),
                type=owner_type,
                typeUrn=owner_type_urn,
            )
        elif isinstance(raw_owner, Urn):
            return models.OwnerClass(
                owner=str(raw_owner),
                type=owner_type,
                typeUrn=owner_type_urn,
            )
        else:
            raise SdkUsageError(
                f"Invalid owner {owner}: {type(owner)} is not a valid owner type"
            )

    # TODO: Return a custom type with deserialized urns, instead of the raw aspect.
    # Ideally we'd also use first-class ownership type urns here, not strings.
    @property
    def owners(self) -> Optional[List[models.OwnerClass]]:
        if owners_aspect := self._get_aspect(models.OwnershipClass):
            return owners_aspect.owners
        return None

    def set_owners(self, owners: OwnersInputType) -> None:
        # TODO: add docs on the default parsing + default ownership type
        parsed_owners = [self._parse_owner_class(owner) for owner in owners]
        self._set_aspect(models.OwnershipClass(owners=parsed_owners))


ContainerInputType: TypeAlias = Union["Container", ContainerKey]


class HasContainer(Entity):
    __slots__ = ()

    def _set_container(self, container: Optional[ContainerInputType]) -> None:
        # We need to allow container to be None. It won't happen for datasets much, but
        # will be required for root containers.
        from datahub.sdk.container import Container

        browse_path: List[Union[str, models.BrowsePathEntryClass]] = []
        if isinstance(container, Container):
            container_urn = container.urn.urn()

            parent_browse_path = container._get_aspect(models.BrowsePathsV2Class)
            if parent_browse_path is None:
                raise SdkUsageError(
                    "Parent container does not have a browse path, so cannot generate one for its children."
                )
            browse_path = [
                *parent_browse_path.path,
                models.BrowsePathEntryClass(
                    id=container_urn,
                    urn=container_urn,
                ),
            ]
        elif container is not None:
            container_urn = container.as_urn()

            browse_path_reversed = [container_urn]
            parent_key = container.parent_key()
            while parent_key is not None:
                browse_path_reversed.append(parent_key.as_urn())
                parent_key = parent_key.parent_key()
            if container.instance is not None:
                browse_path_reversed.append(
                    DataPlatformInstanceUrn(
                        container.platform, container.instance
                    ).urn()
                )

            browse_path = list(reversed(browse_path_reversed))
        else:
            container_urn = None
            browse_path = []

        if container_urn:
            self._set_aspect(models.ContainerClass(container=container_urn))

        self._set_aspect(
            models.BrowsePathsV2Class(
                path=[
                    (
                        entry
                        if isinstance(entry, models.BrowsePathEntryClass)
                        else models.BrowsePathEntryClass(
                            id=entry,
                            urn=entry,
                        )
                    )
                    for entry in browse_path
                ]
            )
        )


TagInputType: TypeAlias = Union[str, TagUrn, models.TagAssociationClass]
TagsInputType: TypeAlias = List[TagInputType]


class HasTags(Entity):
    __slots__ = ()

    # TODO: Return a custom type with deserialized urns, instead of the raw aspect.
    @property
    def tags(self) -> Optional[List[models.TagAssociationClass]]:
        if tags := self._get_aspect(models.GlobalTagsClass):
            return tags.tags
        return None

    @classmethod
    def _parse_tag_association_class(
        cls, tag: TagInputType
    ) -> models.TagAssociationClass:
        if isinstance(tag, models.TagAssociationClass):
            return tag
        elif isinstance(tag, str):
            assert TagUrn.from_string(tag)
        return models.TagAssociationClass(tag=str(tag))

    def set_tags(self, tags: TagsInputType) -> None:
        self._set_aspect(
            models.GlobalTagsClass(
                tags=[self._parse_tag_association_class(tag) for tag in tags]
            )
        )


TermInputType: TypeAlias = Union[
    str, GlossaryTermUrn, models.GlossaryTermAssociationClass
]
TermsInputType: TypeAlias = List[TermInputType]


class HasTerms(Entity):
    __slots__ = ()

    # TODO: Return a custom type with deserialized urns, instead of the raw aspect.
    @property
    def terms(self) -> Optional[List[models.GlossaryTermAssociationClass]]:
        if glossary_terms := self._get_aspect(models.GlossaryTermsClass):
            return glossary_terms.terms
        return None

    @classmethod
    def _parse_glossary_term_association_class(
        cls, term: TermInputType
    ) -> models.GlossaryTermAssociationClass:
        if isinstance(term, models.GlossaryTermAssociationClass):
            return term
        elif isinstance(term, str):
            assert GlossaryTermUrn.from_string(term)
        return models.GlossaryTermAssociationClass(urn=str(term))

    @classmethod
    def _terms_audit_stamp(self) -> models.AuditStampClass:
        return models.AuditStampClass(
            time=0,
            # TODO figure out what to put here
            actor=CorpUserUrn("__ingestion").urn(),
        )

    def set_terms(self, terms: TermsInputType) -> None:
        self._set_aspect(
            models.GlossaryTermsClass(
                terms=[
                    self._parse_glossary_term_association_class(term) for term in terms
                ],
                auditStamp=self._terms_audit_stamp(),
            )
        )


DomainInputType: TypeAlias = Union[str, DomainUrn]


class HasDomain(Entity):
    __slots__ = ()

    @property
    def domain(self) -> Optional[DomainUrn]:
        if domains := self._get_aspect(models.DomainsClass):
            if len(domains.domains) > 1:
                raise SdkUsageError(
                    f"The entity has multiple domains set, but only one is supported: {domains.domains}"
                )
            elif domains.domains:
                domain_str = domains.domains[0]
                return DomainUrn.from_string(domain_str)

        return None

    def set_domain(self, domain: DomainInputType) -> None:
        domain_urn = DomainUrn.from_string(domain)  # basically a type assertion
        self._set_aspect(models.DomainsClass(domains=[str(domain_urn)]))



import logging
import time
import uuid
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple, Type, Union

from datahub._codegen.aspect import _Aspect
from datahub.emitter.mce_builder import (
    make_dataplatform_instance_urn,
    make_domain_urn,
    make_group_urn,
    make_user_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import ContainerKey
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.dremio.dremio_entities import (
    DremioContainer,
    DremioDataset,
    DremioDatasetColumn,
    DremioDatasetType,
    DremioGlossaryTerm,
)
from datahub.metadata.schema_classes import (
    ArrayTypeClass,
    AuditStampClass,
    BooleanTypeClass,
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    BytesTypeClass,
    ContainerClass,
    ContainerPropertiesClass,
    DataPlatformInstanceClass,
    DatasetFieldProfileClass,
    DatasetProfileClass,
    DatasetPropertiesClass,
    DateTypeClass,
    DomainsClass,
    GlossaryTermAssociationClass,
    GlossaryTermInfoClass,
    GlossaryTermsClass,
    MySqlDDLClass,
    NullTypeClass,
    NumberTypeClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    QuantileClass,
    RecordTypeClass,
    SchemaFieldClass,
    SchemaFieldDataTypeClass,
    SchemaMetadataClass,
    StatusClass,
    StringTypeClass,
    SubTypesClass,
    TimeStampClass,
    TimeTypeClass,
    ViewPropertiesClass,
)

logger = logging.getLogger(__name__)
namespace = uuid.NAMESPACE_DNS


class DremioContainerKey(ContainerKey):
    key: str


class SchemaFieldTypeMapper:
    # From https://docs.dremio.com/cloud/reference/sql/data-types/

    FIELD_TYPE_MAPPING: Dict[str, Type] = {
        # Bool
        "boolean": BooleanTypeClass,
        # Binary
        "binary varying": BytesTypeClass,
        # Numbers
        "decimal": NumberTypeClass,
        "integer": NumberTypeClass,
        "bigint": NumberTypeClass,
        "float": NumberTypeClass,
        "double": NumberTypeClass,
        # Dates and times
        "timestamp": DateTypeClass,
        "date": DateTypeClass,
        "time": TimeTypeClass,
        # Strings
        "char": StringTypeClass,
        "character": StringTypeClass,
        "character varying": StringTypeClass,
        # Records
        "row": RecordTypeClass,
        "struct": RecordTypeClass,
        "list": RecordTypeClass,
        "map": RecordTypeClass,
        # Arrays
        "array": ArrayTypeClass,
    }

    @classmethod
    def get_field_type(
        cls, data_type: str, data_size: Optional[int] = None
    ) -> Tuple["SchemaFieldDataTypeClass", str]:
        """
        Maps a Dremio data type and size to a DataHub SchemaFieldDataTypeClass and native data type string.

        :param data_type: The data type string from Dremio.
        :param data_size: The size of the data type, if applicable.
        :return: A tuple containing a SchemaFieldDataTypeClass instance and the native data type string.
        """
        if not data_type:
            logger.warning("Empty data_type provided, defaulting to NullTypeClass.")
            type_class = NullTypeClass
            native_data_type = "NULL"
        else:
            data_type = data_type.lower()
            type_class = cls.FIELD_TYPE_MAPPING.get(data_type, NullTypeClass)

            if data_size:
                native_data_type = f"{data_type}({data_size})"
            else:
                native_data_type = data_type

        try:
            schema_field_type = SchemaFieldDataTypeClass(type=type_class())
            logger.debug(
                f"Mapped data_type '{data_type}' with size '{data_size}' to type class "
                f"'{type_class.__name__}' and native data type '{native_data_type}'."
            )
        except Exception as e:
            logger.error(
                f"Error initializing SchemaFieldDataTypeClass with type '{type_class.__name__}': {e}"
            )
            schema_field_type = SchemaFieldDataTypeClass(type=NullTypeClass())

        return schema_field_type, native_data_type


class DremioAspects:
    def __init__(
        self,
        platform: str,
        ui_url: str,
        env: str,
        ingest_owner: bool,
        domain: Optional[str] = None,
        platform_instance: Optional[str] = None,
    ):
        self.platform = platform
        self.platform_instance = platform_instance
        self.env = env
        self.domain = domain
        self.ui_url = ui_url
        self.ingest_owner = ingest_owner

    def get_container_key(
        self, name: Optional[str], path: Optional[List[str]]
    ) -> DremioContainerKey:
        key = name
        if path:
            key = ".".join(path) + "." + name if name else ".".join(path)

        return DremioContainerKey(
            platform=self.platform,
            instance=self.platform_instance,
            env=str(self.env),
            key=key,
        )

    def get_container_urn(
        self, name: Optional[str] = None, path: Optional[List[str]] = []
    ) -> str:
        container_key = self.get_container_key(name, path)
        return container_key.as_urn()

    def create_domain_aspect(self) -> Optional[_Aspect]:
        if self.domain:
            if self.domain.startswith("urn:li:domain:"):
                return DomainsClass(domains=[self.domain])
            return DomainsClass(
                domains=[
                    make_domain_urn(
                        str(uuid.uuid5(namespace, self.domain)),
                    )
                ]
            )
        return None

    def populate_container_mcp(
        self, container_urn: str, container: DremioContainer
    ) -> Iterable[MetadataWorkUnit]:
        # Container Properties
        container_properties = self._create_container_properties(container)
        mcp = MetadataChangeProposalWrapper(
            entityUrn=container_urn,
            aspect=container_properties,
            aspectName=ContainerPropertiesClass.ASPECT_NAME,
        )
        yield mcp.as_workunit()

        if not container.path:
            browse_paths_v2 = self._create_browse_paths_containers(container)
            if browse_paths_v2:
                mcp = MetadataChangeProposalWrapper(
                    entityUrn=container_urn,
                    aspect=browse_paths_v2,
                )
                yield mcp.as_workunit()

        # Container Class Folders
        container_class = self._create_container_class(container)
        if container_class:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=container_urn,
                aspect=container_class,
            )
            yield mcp.as_workunit()

        # Data Platform Instance
        data_platform_instance = self._create_data_platform_instance()
        if data_platform_instance:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=container_urn,
                aspect=data_platform_instance,
            )
            yield mcp.as_workunit()

        # SubTypes
        subtypes = SubTypesClass(typeNames=[container.subclass])
        mcp = MetadataChangeProposalWrapper(
            entityUrn=container_urn,
            aspect=subtypes,
        )
        yield mcp.as_workunit()

        # Status
        status = StatusClass(removed=False)
        mcp = MetadataChangeProposalWrapper(
            entityUrn=container_urn,
            aspect=status,
        )
        yield mcp.as_workunit()

    def populate_dataset_mcp(
        self, dataset_urn: str, dataset: DremioDataset
    ) -> Iterable[MetadataWorkUnit]:
        # Dataset Properties
        dataset_properties = self._create_dataset_properties(dataset)
        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=dataset_properties,
        )
        yield mcp.as_workunit()

        # Ownership
        ownership = self._create_ownership(dataset)
        if ownership:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=ownership,
            )
            yield mcp.as_workunit()

        # SubTypes
        subtypes = SubTypesClass(typeNames=[dataset.dataset_type.value])
        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=subtypes,
        )
        yield mcp.as_workunit()

        # Data Platform Instance
        data_platform_instance = self._create_data_platform_instance()
        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=data_platform_instance,
        )
        yield mcp.as_workunit()

        # Container Class
        container_class = self._create_container_class(dataset)
        if container_class:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=container_class,
            )
            yield mcp.as_workunit()

        # View Definition
        if dataset.dataset_type == DremioDatasetType.VIEW:
            view_definition = self._create_view_properties(dataset)
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=view_definition,
            )
            yield mcp.as_workunit()

        # Glossary Terms
        if dataset.glossary_terms:
            glossary_terms = self._create_glossary_terms(dataset)
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=glossary_terms,
            )
            yield mcp.as_workunit()

        # Schema Metadata
        if dataset.columns:
            schema_metadata = self._create_schema_metadata(dataset)
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=schema_metadata,
            )
            yield mcp.as_workunit()

        else:
            logger.warning(
                f"Dataset {dataset.path}.{dataset.resource_name} has not been queried in Dremio"
            )
            logger.warning(
                f"Dataset {dataset.path}.{dataset.resource_name} will have a null schema"
            )

        # Status
        status = StatusClass(removed=False)
        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=status,
        )
        yield mcp.as_workunit()

    def populate_glossary_term_mcp(
        self, glossary_term: DremioGlossaryTerm
    ) -> Iterable[MetadataWorkUnit]:
        glossary_term_info = self._create_glossary_term_info(glossary_term)
        mcp = MetadataChangeProposalWrapper(
            entityUrn=glossary_term.urn,
            aspect=glossary_term_info,
        )
        yield mcp.as_workunit()

    def populate_profile_aspect(self, profile_data: Dict) -> DatasetProfileClass:
        field_profiles = [
            self._create_field_profile(field_name, field_stats)
            for field_name, field_stats in profile_data.get("column_stats", {}).items()
        ]
        return DatasetProfileClass(
            timestampMillis=round(time.time() * 1000),
            rowCount=profile_data.get("row_count"),
            columnCount=profile_data.get("column_count"),
            fieldProfiles=field_profiles,
        )

    def _create_container_properties(
        self, container: DremioContainer
    ) -> ContainerPropertiesClass:
        return ContainerPropertiesClass(
            name=container.container_name,
            qualifiedName=f"{'.'.join(container.path) + '.' if container.path else ''}{container.container_name}",
            description=container.description,
            env=self.env,
        )

    def _create_browse_paths_containers(
        self, entity: DremioContainer
    ) -> Optional[BrowsePathsV2Class]:
        paths = []

        if entity.subclass == "Dremio Space":
            paths.append(BrowsePathEntryClass(id="Spaces"))
        elif entity.subclass == "Dremio Source":
            paths.append(BrowsePathEntryClass(id="Sources"))
        if paths:
            return BrowsePathsV2Class(path=paths)
        return None

    def _create_container_class(
        self, entity: Union[DremioContainer, DremioDataset]
    ) -> Optional[ContainerClass]:
        if entity.path:
            return ContainerClass(container=self.get_container_urn(path=entity.path))
        return None

    def _create_data_platform_instance(self) -> DataPlatformInstanceClass:
        return DataPlatformInstanceClass(
            platform=f"urn:li:dataPlatform:{self.platform}",
            instance=(
                make_dataplatform_instance_urn(self.platform, self.platform_instance)
                if self.platform_instance
                else None
            ),
        )

    def _create_dataset_properties(
        self, dataset: DremioDataset
    ) -> DatasetPropertiesClass:
        return DatasetPropertiesClass(
            name=dataset.resource_name,
            qualifiedName=f"{'.'.join(dataset.path)}.{dataset.resource_name}",
            description=dataset.description,
            externalUrl=self._create_external_url(dataset=dataset),
            created=TimeStampClass(
                time=round(
                    datetime.strptime(
                        dataset.created, "%Y-%m-%d %H:%M:%S.%f"
                    ).timestamp()
                    * 1000
                )
                if hasattr(dataset, "created")
                else 0,
            ),
        )

    def _create_external_url(self, dataset: DremioDataset) -> str:
        container_type = "source"
        dataset_url_path = '"' + dataset.path[0] + '"/'

        if len(dataset.path) > 1:
            dataset_url_path = (
                dataset_url_path + '"' + '"."'.join(dataset.path[1:]) + '".'
            )

        if dataset.dataset_type == DremioDatasetType.VIEW:
            container_type = "space"
        elif dataset.path[0].startswith("@"):
            container_type = "home"

        return f'{self.ui_url}/{container_type}/{dataset_url_path}"{dataset.resource_name}"'

    def _create_ownership(self, dataset: DremioDataset) -> Optional[OwnershipClass]:
        if self.ingest_owner and dataset.owner:
            owner_urn = (
                make_user_urn(dataset.owner)
                if dataset.owner_type == "USER"
                else make_group_urn(dataset.owner)
            )
            ownership: OwnershipClass = OwnershipClass(
                owners=[
                    OwnerClass(
                        owner=owner_urn,
                        type=OwnershipTypeClass.TECHNICAL_OWNER,
                    )
                ]
            )
            return ownership

        return None

    def _create_glossary_terms(self, entity: DremioDataset) -> GlossaryTermsClass:
        return GlossaryTermsClass(
            terms=[
                GlossaryTermAssociationClass(urn=term.urn)
                for term in entity.glossary_terms
            ],
            auditStamp=AuditStampClass(
                time=round(time.time() * 1000),
                actor="urn:li:corpuser:admin",
            ),
        )

    def _create_schema_metadata(self, dataset: DremioDataset) -> SchemaMetadataClass:
        return SchemaMetadataClass(
            schemaName=f"{'.'.join(dataset.path)}.{dataset.resource_name}",
            platform=f"urn:li:dataPlatform:{self.platform}",
            version=0,
            fields=[self._create_schema_field(column) for column in dataset.columns],
            platformSchema=MySqlDDLClass(""),
            hash="",
        )

    def _create_schema_field(self, column: DremioDatasetColumn) -> SchemaFieldClass:
        type_class, native_data_type = SchemaFieldTypeMapper.get_field_type(
            column.data_type,
            column.column_size,
        )
        return SchemaFieldClass(
            fieldPath=column.name,
            type=type_class,
            nativeDataType=native_data_type,
            nullable=column.is_nullable == "YES",
        )

    def _create_view_properties(
        self, dataset: DremioDataset
    ) -> Optional[ViewPropertiesClass]:
        if not dataset.sql_definition:
            return None
        return ViewPropertiesClass(
            materialized=False,
            viewLanguage="SQL",
            viewLogic=dataset.sql_definition,
        )

    def _create_glossary_term_info(
        self, glossary_term: DremioGlossaryTerm
    ) -> GlossaryTermInfoClass:
        return GlossaryTermInfoClass(
            definition="",
            termSource=self.platform,
            name=glossary_term.glossary_term,
        )

    def _create_field_profile(
        self, field_name: str, field_stats: Dict
    ) -> DatasetFieldProfileClass:
        quantiles = field_stats.get("quantiles")
        return DatasetFieldProfileClass(
            fieldPath=field_name,
            uniqueCount=field_stats.get("distinct_count"),
            nullCount=field_stats.get("null_count"),
            min=str(field_stats.get("min")) if field_stats.get("min") else None,
            max=str(field_stats.get("max")) if field_stats.get("max") else None,
            mean=str(field_stats.get("mean")) if field_stats.get("mean") else None,
            median=str(field_stats.get("median"))
            if field_stats.get("median")
            else None,
            stdev=str(field_stats.get("stdev")) if field_stats.get("stdev") else None,
            quantiles=[
                QuantileClass(quantile=str(0.25), value=str(quantiles[0])),
                QuantileClass(quantile=str(0.75), value=str(quantiles[1])),
            ]
            if quantiles
            else None,
            sampleValues=field_stats.get("sample_values"),
        )



import abc
from typing import List, Optional, Type, Union

from typing_extensions import Self

import datahub.metadata.schema_classes as models
from datahub.emitter.mce_builder import Aspect as AspectTypeVar
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.errors import SdkUsageError
from datahub.metadata.urns import Urn
from datahub.utilities.urns._urn_base import _SpecificUrn


class Entity:
    __slots__ = ("_urn", "_prev_aspects", "_aspects")

    def __init__(self, /, urn: Urn):
        # This method is not meant for direct usage.
        if type(self) is Entity:
            raise SdkUsageError(f"{Entity.__name__} cannot be instantiated directly.")

        assert isinstance(urn, self.get_urn_type())
        self._urn: _SpecificUrn = urn

        # prev_aspects is None means this was created from scratch
        self._prev_aspects: Optional[models.AspectBag] = None
        self._aspects: models.AspectBag = {}

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        # If an init method from a subclass adds required fields, it also needs to override this method.
        # An alternative approach would call cls.__new__() to bypass the init method, but it's a bit
        # too hacky for my taste.
        entity = cls(urn=urn)
        return entity._init_from_graph(current_aspects)

    def _init_from_graph(self, current_aspects: models.AspectBag) -> Self:
        self._prev_aspects = current_aspects
        aspect: models._Aspect
        for aspect_name, aspect in (current_aspects or {}).items():  # type: ignore
            aspect_copy = type(aspect).from_obj(aspect.to_obj())
            self._aspects[aspect_name] = aspect_copy  # type: ignore
        return self

    @classmethod
    @abc.abstractmethod
    def get_urn_type(cls) -> Type[_SpecificUrn]: ...

    @property
    def urn(self) -> _SpecificUrn:
        return self._urn

    def _get_aspect(
        self,
        aspect_type: Type[AspectTypeVar],
        /,
    ) -> Optional[AspectTypeVar]:
        return self._aspects.get(aspect_type.ASPECT_NAME)  # type: ignore

    def _set_aspect(self, value: AspectTypeVar, /) -> None:
        self._aspects[value.ASPECT_NAME] = value  # type: ignore

    def _setdefault_aspect(self, default_aspect: AspectTypeVar, /) -> AspectTypeVar:
        # Similar semantics to dict.setdefault.
        if existing_aspect := self._get_aspect(type(default_aspect)):
            return existing_aspect
        self._set_aspect(default_aspect)
        return default_aspect

    def _as_mcps(
        self,
        change_type: Union[str, models.ChangeTypeClass] = models.ChangeTypeClass.UPSERT,
    ) -> List[MetadataChangeProposalWrapper]:
        urn_str = str(self.urn)

        mcps = []
        for aspect in self._aspects.values():
            assert isinstance(aspect, models._Aspect)
            mcps.append(
                MetadataChangeProposalWrapper(
                    entityUrn=urn_str,
                    aspect=aspect,
                    changeType=change_type,
                )
            )
        return mcps

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.urn}')"



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import TestDefinitionClass
from .....schema_classes import TestDefinitionTypeClass
from .....schema_classes import TestInfoClass
from .....schema_classes import TestResultClass
from .....schema_classes import TestResultTypeClass
from .....schema_classes import TestResultsClass


TestDefinition = TestDefinitionClass
TestDefinitionType = TestDefinitionTypeClass
TestInfo = TestInfoClass
TestResult = TestResultClass
TestResultType = TestResultTypeClass
TestResults = TestResultsClass

# fmt: on



import logging
from pathlib import Path

import click
from click_default_group import DefaultGroup

from datahub.api.entities.corpgroup.corpgroup import (
    CorpGroup,
    CorpGroupGenerationConfig,
)
from datahub.cli.specific.file_loader import load_file
from datahub.ingestion.graph.client import get_default_graph
from datahub.telemetry import telemetry
from datahub.upgrade import upgrade

logger = logging.getLogger(__name__)


@click.group(cls=DefaultGroup, default="upsert")
def group() -> None:
    """A group of commands to interact with the Group entity in DataHub."""
    pass


@group.command(
    name="upsert",
)
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@click.option(
    "--override-editable",
    required=False,
    is_flag=True,
    default=False,
    help="When set, writes to the editable section of the metadata graph, overwriting writes from the UI",
)
@upgrade.check_upgrade
@telemetry.with_telemetry()
def upsert(file: Path, override_editable: bool) -> None:
    """Create or Update a Group with embedded Users"""

    config_dict = load_file(file)
    group_configs = config_dict if isinstance(config_dict, list) else [config_dict]
    with get_default_graph() as emitter:
        for group_config in group_configs:
            try:
                datahub_group = CorpGroup.parse_obj(group_config)
                for mcp in datahub_group.generate_mcp(
                    generation_config=CorpGroupGenerationConfig(
                        override_editable=override_editable, datahub_graph=emitter
                    )
                ):
                    emitter.emit(mcp)
                click.secho(
                    f"Update succeeded for group {datahub_group.urn}.", fg="green"
                )
            except Exception as e:
                click.secho(
                    f"Update failed for id {group_config.get('id')}. due to {e}",
                    fg="red",
                )



import json
import logging
import time
from dataclasses import dataclass, field as dataclass_field
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional

import requests
from pydantic import Field, validator
from simple_salesforce import Salesforce
from simple_salesforce.exceptions import SalesforceAuthenticationFailed

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import (
    AllowDenyPattern,
    ConfigModel,
    ConfigurationError,
)
from datahub.configuration.source_common import DatasetSourceConfigMixin
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import add_domain_to_entity_wu
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import DatasetSubTypes
from datahub.ingestion.source_config.operation_config import (
    OperationConfig,
    is_profiling_enabled,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BooleanTypeClass,
    BytesTypeClass,
    DataPlatformInstanceClass,
    DatasetProfileClass,
    DatasetPropertiesClass,
    DateTypeClass,
    EnumTypeClass,
    ForeignKeyConstraintClass,
    GlobalTagsClass,
    NullTypeClass,
    NumberTypeClass,
    OperationClass,
    OperationTypeClass,
    OtherSchemaClass,
    RecordTypeClass,
    SchemaFieldClass,
    SchemaFieldDataTypeClass,
    SchemaMetadataClass,
    StringTypeClass,
    SubTypesClass,
    TagAssociationClass,
)
from datahub.utilities import config_clean
from datahub.utilities.lossy_collections import LossyList

logger = logging.getLogger(__name__)


class SalesforceAuthType(Enum):
    USERNAME_PASSWORD = "USERNAME_PASSWORD"
    DIRECT_ACCESS_TOKEN = "DIRECT_ACCESS_TOKEN"
    JSON_WEB_TOKEN = "JSON_WEB_TOKEN"


class SalesforceProfilingConfig(ConfigModel):
    enabled: bool = Field(
        default=False,
        description="Whether profiling should be done. Supports only table-level profiling at this stage",
    )
    operation_config: OperationConfig = Field(
        default_factory=OperationConfig,
        description="Experimental feature. To specify operation configs.",
    )

    # TODO - support field level profiling


class SalesforceConfig(DatasetSourceConfigMixin):
    platform: str = "salesforce"

    auth: SalesforceAuthType = SalesforceAuthType.USERNAME_PASSWORD

    # Username, Password Auth
    username: Optional[str] = Field(description="Salesforce username")
    password: Optional[str] = Field(description="Password for Salesforce user")
    consumer_key: Optional[str] = Field(
        description="Consumer key for Salesforce JSON web token access"
    )
    private_key: Optional[str] = Field(
        description="Private key as a string for Salesforce JSON web token access"
    )
    security_token: Optional[str] = Field(
        description="Security token for Salesforce username"
    )
    # client_id, client_secret not required

    # Direct - Instance URL, Access Token Auth
    instance_url: Optional[str] = Field(
        description="Salesforce instance url. e.g. https://MyDomainName.my.salesforce.com"
    )
    # Flag to indicate whether the instance is production or sandbox
    is_sandbox: bool = Field(
        default=False, description="Connect to Sandbox instance of your Salesforce"
    )
    access_token: Optional[str] = Field(description="Access token for instance url")

    ingest_tags: Optional[bool] = Field(
        default=False,
        description="Ingest Tags from source. This will override Tags entered from UI",
    )

    object_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for Salesforce objects to filter in ingestion.",
    )
    domain: Dict[str, AllowDenyPattern] = Field(
        default=dict(),
        description='Regex patterns for tables/schemas to describe domain_key domain key (domain_key can be any string like "sales".) There can be multiple domain keys specified.',
    )
    api_version: Optional[str] = Field(
        description="If specified, overrides default version used by the Salesforce package. Example value: '59.0'"
    )

    profiling: SalesforceProfilingConfig = SalesforceProfilingConfig()

    profile_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for profiles to filter in ingestion, allowed by the `object_pattern`.",
    )

    def is_profiling_enabled(self) -> bool:
        return self.profiling.enabled and is_profiling_enabled(
            self.profiling.operation_config
        )

    @validator("instance_url")
    def remove_trailing_slash(cls, v):
        return config_clean.remove_trailing_slashes(v)


@dataclass
class SalesforceSourceReport(SourceReport):
    filtered: LossyList[str] = dataclass_field(default_factory=LossyList)

    def report_dropped(self, ent_name: str) -> None:
        self.filtered.append(ent_name)


# https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_describesObjects_describesObjectresult.htm#FieldType
FIELD_TYPE_MAPPING = {
    "string": StringTypeClass,
    "boolean": BooleanTypeClass,
    "int": NumberTypeClass,
    "integer": NumberTypeClass,
    "long": NumberTypeClass,
    "double": NumberTypeClass,
    "date": DateTypeClass,
    "datetime": DateTypeClass,
    "time": DateTypeClass,
    "id": StringTypeClass,  # Primary Key
    "picklist": EnumTypeClass,
    "address": RecordTypeClass,
    "location": RecordTypeClass,
    "reference": StringTypeClass,  # Foreign Key
    "currency": NumberTypeClass,
    "textarea": StringTypeClass,
    "percent": NumberTypeClass,
    "phone": StringTypeClass,
    "url": StringTypeClass,
    "email": StringTypeClass,
    "combobox": StringTypeClass,
    "multipicklist": StringTypeClass,
    "base64": BytesTypeClass,
    "anyType": NullTypeClass,
    "encryptedstring": StringTypeClass,
}


@platform_name("Salesforce")
@config_class(SalesforceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(
    capability_name=SourceCapability.PLATFORM_INSTANCE,
    description="Can be equivalent to Salesforce organization",
)
@capability(
    capability_name=SourceCapability.DOMAINS,
    description="Supported via the `domain` config field",
)
@capability(
    capability_name=SourceCapability.DATA_PROFILING,
    description="Only table level profiling is supported via `profiling.enabled` config field",
)
@capability(
    capability_name=SourceCapability.DELETION_DETECTION,
    description="Not supported yet",
    supported=False,
)
@capability(
    capability_name=SourceCapability.SCHEMA_METADATA,
    description="Enabled by default",
)
@capability(
    capability_name=SourceCapability.TAGS,
    description="Enabled by default",
)
class SalesforceSource(Source):
    base_url: str
    config: SalesforceConfig
    report: SalesforceSourceReport
    session: requests.Session
    sf: Salesforce
    fieldCounts: Dict[str, int]

    def __init__(self, config: SalesforceConfig, ctx: PipelineContext) -> None:
        super().__init__(ctx)
        self.config = config
        self.report = SalesforceSourceReport()
        self.session = requests.Session()
        self.platform: str = "salesforce"
        self.fieldCounts = {}
        common_args: Dict[str, Any] = {
            "domain": "test" if self.config.is_sandbox else None,
            "session": self.session,
        }
        if self.config.api_version:
            common_args["version"] = self.config.api_version

        try:
            if self.config.auth is SalesforceAuthType.DIRECT_ACCESS_TOKEN:
                logger.debug("Access Token Provided in Config")
                assert self.config.access_token is not None, (
                    "Config access_token is required for DIRECT_ACCESS_TOKEN auth"
                )
                assert self.config.instance_url is not None, (
                    "Config instance_url is required for DIRECT_ACCESS_TOKEN auth"
                )

                self.sf = Salesforce(
                    instance_url=self.config.instance_url,
                    session_id=self.config.access_token,
                    **common_args,
                )
            elif self.config.auth is SalesforceAuthType.USERNAME_PASSWORD:
                logger.debug("Username/Password Provided in Config")
                assert self.config.username is not None, (
                    "Config username is required for USERNAME_PASSWORD auth"
                )
                assert self.config.password is not None, (
                    "Config password is required for USERNAME_PASSWORD auth"
                )
                assert self.config.security_token is not None, (
                    "Config security_token is required for USERNAME_PASSWORD auth"
                )

                self.sf = Salesforce(
                    username=self.config.username,
                    password=self.config.password,
                    security_token=self.config.security_token,
                    **common_args,
                )

            elif self.config.auth is SalesforceAuthType.JSON_WEB_TOKEN:
                logger.debug("Json Web Token provided in the config")
                assert self.config.username is not None, (
                    "Config username is required for JSON_WEB_TOKEN auth"
                )
                assert self.config.consumer_key is not None, (
                    "Config consumer_key is required for JSON_WEB_TOKEN auth"
                )
                assert self.config.private_key is not None, (
                    "Config private_key is required for JSON_WEB_TOKEN auth"
                )

                self.sf = Salesforce(
                    username=self.config.username,
                    consumer_key=self.config.consumer_key,
                    privatekey=self.config.private_key,
                    **common_args,
                )

        except SalesforceAuthenticationFailed as e:
            logger.error(e)
            if "API_CURRENTLY_DISABLED" in str(e):
                # https://help.salesforce.com/s/articleView?id=001473830&type=1
                error = "Salesforce login failed. Please make sure user has API Enabled Access."
            else:
                error = "Salesforce login failed. Please verify your credentials."
                if (
                    self.config.instance_url
                    and "sandbox" in self.config.instance_url.lower()
                ):
                    error += "Please set `is_sandbox: True` in recipe if this is sandbox account."
            raise ConfigurationError(error) from e

        if not self.config.api_version:
            # List all REST API versions and use latest one
            versions_url = "https://{instance}/services/data/".format(
                instance=self.sf.sf_instance,
            )
            versions_response = self.sf._call_salesforce("GET", versions_url).json()
            latest_version = versions_response[-1]
            version = latest_version["version"]
            # we could avoid setting the version like below (after the Salesforce object has been already initiated
            # above), since, according to the docs:
            # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_versions.htm
            # we don't need to be authenticated to list the versions (so we could perform this call before even
            # authenticating)
            self.sf.sf_version = version

        self.base_url = "https://{instance}/services/data/v{sf_version}/".format(
            instance=self.sf.sf_instance, sf_version=self.sf.sf_version
        )

        logger.debug(
            "Using Salesforce REST API version: {version}".format(
                version=self.sf.sf_version
            )
        )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        try:
            sObjects = self.get_salesforce_objects()
        except Exception as e:
            if "sObject type 'EntityDefinition' is not supported." in str(e):
                # https://developer.salesforce.com/docs/atlas.en-us.api_tooling.meta/api_tooling/tooling_api_objects_entitydefinition.htm
                raise ConfigurationError(
                    "Salesforce EntityDefinition query failed. "
                    "Please verify if user has 'View Setup and Configuration' permission."
                ) from e
            raise e
        else:
            for sObject in sObjects:
                yield from self.get_salesforce_object_workunits(sObject)

    def get_salesforce_object_workunits(
        self, sObject: dict
    ) -> Iterable[MetadataWorkUnit]:
        sObjectName = sObject["QualifiedApiName"]

        if not self.config.object_pattern.allowed(sObjectName):
            self.report.report_dropped(sObjectName)
            logger.debug(
                "Skipping {sObject}, as it is not allowed by object_pattern".format(
                    sObject=sObjectName
                )
            )
            return

        datasetUrn = builder.make_dataset_urn_with_platform_instance(
            self.platform,
            sObjectName,
            self.config.platform_instance,
            self.config.env,
        )

        customObject = {}
        if sObjectName.endswith("__c"):  # Is Custom Object
            customObject = self.get_custom_object_details(sObject["DeveloperName"])

            # Table Created, LastModified is available for Custom Object
            yield from self.get_operation_workunit(customObject, datasetUrn)

        yield self.get_properties_workunit(sObject, customObject, datasetUrn)

        yield from self.get_schema_metadata_workunit(
            sObjectName, sObject, customObject, datasetUrn
        )

        yield self.get_subtypes_workunit(sObjectName, datasetUrn)

        if self.config.platform_instance is not None:
            yield self.get_platform_instance_workunit(datasetUrn)

        if self.config.domain is not None:
            yield from self.get_domain_workunit(sObjectName, datasetUrn)

        if self.config.is_profiling_enabled() and self.config.profile_pattern.allowed(
            sObjectName
        ):
            yield from self.get_profile_workunit(sObjectName, datasetUrn)

    def get_custom_object_details(self, sObjectDeveloperName: str) -> dict:
        customObject = {}
        query_url = (
            self.base_url
            + "tooling/query/?q=SELECT Description, Language, ManageableState, "
            + "CreatedDate, CreatedBy.Username, LastModifiedDate, LastModifiedBy.Username "
            + f"FROM CustomObject where DeveloperName='{sObjectDeveloperName}'"
        )
        custom_objects_response = self.sf._call_salesforce("GET", query_url).json()
        if len(custom_objects_response["records"]) > 0:
            logger.debug("Salesforce CustomObject query returned with details")
            customObject = custom_objects_response["records"][0]
        return customObject

    def get_salesforce_objects(self) -> List:
        # Using Describe Global REST API returns many more objects than required.
        # Response does not have the attribute ("customizable") that can be used
        # to filter out entities not on ObjectManager UI. Hence SOQL on EntityDefinition
        # object is used instead, as suggested by salesforce support.

        query_url = (
            self.base_url
            + "tooling/query/?q=SELECT DurableId,QualifiedApiName,DeveloperName,"
            + "Label,PluralLabel,InternalSharingModel,ExternalSharingModel,DeploymentStatus "
            + "FROM EntityDefinition WHERE IsCustomizable = true"
        )
        entities_response = self.sf._call_salesforce("GET", query_url).json()
        logger.debug(
            "Salesforce EntityDefinition query returned {count} sObjects".format(
                count=len(entities_response["records"])
            )
        )
        return entities_response["records"]

    def get_domain_workunit(
        self, dataset_name: str, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        domain_urn: Optional[str] = None

        for domain, pattern in self.config.domain.items():
            if pattern.allowed(dataset_name):
                domain_urn = builder.make_domain_urn(domain)

        if domain_urn:
            yield from add_domain_to_entity_wu(
                domain_urn=domain_urn, entity_urn=datasetUrn
            )

    def get_platform_instance_workunit(self, datasetUrn: str) -> MetadataWorkUnit:
        dataPlatformInstance = DataPlatformInstanceClass(
            builder.make_data_platform_urn(self.platform),
            instance=builder.make_dataplatform_instance_urn(
                self.platform,
                self.config.platform_instance,  # type:ignore
            ),
        )

        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=dataPlatformInstance
        ).as_workunit()

    def get_operation_workunit(
        self, customObject: dict, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        reported_time: int = int(time.time() * 1000)

        if customObject.get("CreatedBy") and customObject.get("CreatedDate"):
            timestamp = self.get_time_from_salesforce_timestamp(
                customObject["CreatedDate"]
            )
            operation = OperationClass(
                timestampMillis=reported_time,
                operationType=OperationTypeClass.CREATE,
                lastUpdatedTimestamp=timestamp,
                actor=builder.make_user_urn(customObject["CreatedBy"]["Username"]),
            )

            yield MetadataChangeProposalWrapper(
                entityUrn=datasetUrn, aspect=operation
            ).as_workunit()

            # Note - Object Level LastModified captures changes at table level metadata e.g. table
            # description and does NOT capture field level metadata e.g. new field added, existing
            # field updated

            if customObject.get("LastModifiedBy") and customObject.get(
                "LastModifiedDate"
            ):
                timestamp = self.get_time_from_salesforce_timestamp(
                    customObject["LastModifiedDate"]
                )
                operation = OperationClass(
                    timestampMillis=reported_time,
                    operationType=OperationTypeClass.ALTER,
                    lastUpdatedTimestamp=timestamp,
                    actor=builder.make_user_urn(
                        customObject["LastModifiedBy"]["Username"]
                    ),
                )
                yield MetadataChangeProposalWrapper(
                    entityUrn=datasetUrn, aspect=operation
                ).as_workunit()

    def get_time_from_salesforce_timestamp(self, date: str) -> int:
        return round(
            datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp() * 1000
        )

    def get_properties_workunit(
        self, sObject: dict, customObject: Dict[str, str], datasetUrn: str
    ) -> MetadataWorkUnit:
        propertyLabels = {
            # from EntityDefinition
            "DurableId": "Durable Id",
            "DeveloperName": "Developer Name",
            "QualifiedApiName": "Qualified API Name",
            "Label": "Label",
            "PluralLabel": "Plural Label",
            "InternalSharingModel": "Internal Sharing Model",
            "ExternalSharingModel": "External Sharing Model",
            # from CustomObject
            "ManageableState": "Manageable State",
            "Language": "Language",
        }

        sObjectProperties = {
            propertyLabels[k]: str(v)
            for k, v in sObject.items()
            if k in propertyLabels and v is not None
        }
        sObjectProperties.update(
            {
                propertyLabels[k]: str(v)
                for k, v in customObject.items()
                if k in propertyLabels and v is not None
            }
        )

        datasetProperties = DatasetPropertiesClass(
            name=sObject["Label"],
            description=customObject.get("Description"),
            customProperties=sObjectProperties,
        )
        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=datasetProperties
        ).as_workunit()

    def get_subtypes_workunit(
        self, sObjectName: str, datasetUrn: str
    ) -> MetadataWorkUnit:
        subtypes: List[str] = []
        if sObjectName.endswith("__c"):
            subtypes.append(DatasetSubTypes.SALESFORCE_CUSTOM_OBJECT)
        else:
            subtypes.append(DatasetSubTypes.SALESFORCE_STANDARD_OBJECT)

        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=SubTypesClass(typeNames=subtypes)
        ).as_workunit()

    def get_profile_workunit(
        self, sObjectName: str, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        # Here approximate record counts as returned by recordCount API are used as rowCount
        # In future, count() SOQL query may be used instead, if required, might be more expensive
        sObject_records_count_url = (
            f"{self.base_url}limits/recordCount?sObjects={sObjectName}"
        )

        sObject_record_count_response = self.sf._call_salesforce(
            "GET", sObject_records_count_url
        ).json()

        logger.debug(
            "Received Salesforce {sObject} record count response".format(
                sObject=sObjectName
            )
        )

        for entry in sObject_record_count_response.get("sObjects", []):
            datasetProfile = DatasetProfileClass(
                timestampMillis=int(time.time() * 1000),
                rowCount=entry["count"],
                columnCount=self.fieldCounts[sObjectName],
            )
            yield MetadataChangeProposalWrapper(
                entityUrn=datasetUrn, aspect=datasetProfile
            ).as_workunit()

    # Here field description is created from label, description and inlineHelpText
    def _get_field_description(self, field: dict, customField: dict) -> str:
        if "Label" not in field or field["Label"] is None:
            desc = ""
        elif field["Label"].startswith("#"):
            desc = "\\" + field["Label"]
        else:
            desc = field["Label"]

        text = field.get("FieldDefinition", {}).get("Description", None)
        if text:
            prefix = "\\" if text.startswith("#") else ""
            desc += f"\n\n{prefix}{text}"

        text = field.get("InlineHelpText", None)
        if text:
            prefix = "\\" if text.startswith("#") else ""
            desc += f"\n\n{prefix}{text}"

        return desc

    # Here jsonProps is used to add additional salesforce field level properties.
    def _get_field_json_props(self, field: dict, customField: dict) -> str:
        jsonProps = {}

        if field.get("IsUnique"):
            jsonProps["IsUnique"] = True

        return json.dumps(jsonProps)

    def _get_schema_field(
        self,
        sObjectName: str,
        fieldName: str,
        fieldType: str,
        field: dict,
        customField: dict,
    ) -> SchemaFieldClass:
        fieldPath = fieldName

        TypeClass = FIELD_TYPE_MAPPING.get(fieldType)
        if TypeClass is None:
            self.report.warning(
                message="Unable to map field type to metadata schema",
                context=f"{fieldType} for {fieldName} of {sObjectName}",
            )
            TypeClass = NullTypeClass

        fieldTags: List[str] = self.get_field_tags(fieldName, field)

        description = self._get_field_description(field, customField)

        schemaField = SchemaFieldClass(
            fieldPath=fieldPath,
            type=SchemaFieldDataTypeClass(type=TypeClass()),  # type:ignore
            description=description,
            # nativeDataType is set to data type shown on salesforce user interface,
            # not the corresponding API data type names.
            nativeDataType=field["FieldDefinition"]["DataType"],
            nullable=field["IsNillable"],
            globalTags=get_tags(fieldTags) if self.config.ingest_tags else None,
            jsonProps=self._get_field_json_props(field, customField),
        )

        # Created and LastModified Date and Actor are available for Custom Fields only
        if customField.get("CreatedDate") and customField.get("CreatedBy"):
            schemaField.created = self.get_audit_stamp(
                customField["CreatedDate"], customField["CreatedBy"]["Username"]
            )
        if customField.get("LastModifiedDate") and customField.get("LastModifiedBy"):
            schemaField.lastModified = self.get_audit_stamp(
                customField["LastModifiedDate"],
                customField["LastModifiedBy"]["Username"],
            )

        return schemaField

    def get_field_tags(self, fieldName: str, field: dict) -> List[str]:
        fieldTags: List[str] = []

        if fieldName.endswith("__c"):
            fieldTags.append("Custom")

        # https://developer.salesforce.com/docs/atlas.en-us.object_reference.meta/object_reference/system_fields.htm
        sfSystemFields = [
            "Id",
            "IsDeleted",
            "CreatedById",
            "CreatedDate",
            "LastModifiedById",
            "LastModifiedDate",
            "SystemModstamp",
        ]

        if fieldName in sfSystemFields:
            fieldTags.append("SystemField")

        if field["FieldDefinition"]["ComplianceGroup"] is not None:
            # CCPA, COPPA, GDPR, HIPAA, PCI, PersonalInfo, PII
            fieldTags.extend(
                iter(field["FieldDefinition"]["ComplianceGroup"].split(";"))
            )
        return fieldTags

    def get_audit_stamp(self, date: str, username: str) -> AuditStampClass:
        return AuditStampClass(
            time=self.get_time_from_salesforce_timestamp(date),
            actor=builder.make_user_urn(username),
        )

    def get_schema_metadata_workunit(
        self, sObjectName: str, sObject: dict, customObject: dict, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        sObject_fields_query_url = (
            self.base_url
            + "tooling/query?q=SELECT "
            + "QualifiedApiName,DeveloperName,Label, FieldDefinition.DataType, DataType,"
            + "FieldDefinition.LastModifiedDate, FieldDefinition.LastModifiedBy.Username,"
            + "Precision, Scale, Length, Digits ,FieldDefinition.IsIndexed, IsUnique,"
            + "IsCompound, IsComponent, ReferenceTo, FieldDefinition.ComplianceGroup,"
            + "RelationshipName, IsNillable, FieldDefinition.Description, InlineHelpText "
            + "FROM EntityParticle WHERE EntityDefinitionId='{}'".format(
                sObject["DurableId"]
            )
        )

        sObject_fields_response = self.sf._call_salesforce(
            "GET", sObject_fields_query_url
        ).json()

        logger.debug(f"Received Salesforce {sObjectName} fields response")

        sObject_custom_fields_query_url = (
            self.base_url
            + "tooling/query?q=SELECT "
            + "DeveloperName,CreatedDate,CreatedBy.Username,InlineHelpText,"
            + "LastModifiedDate,LastModifiedBy.Username "
            + "FROM CustomField WHERE EntityDefinitionId='{}'".format(
                sObject["DurableId"]
            )
        )

        customFields: Dict[str, Dict] = {}
        try:
            sObject_custom_fields_response = self.sf._call_salesforce(
                "GET", sObject_custom_fields_query_url
            ).json()

            logger.debug(
                "Received Salesforce {sObject} custom fields response".format(
                    sObject=sObjectName
                )
            )

        except Exception as e:
            error = "Salesforce CustomField query failed. "
            if "sObject type 'CustomField' is not supported." in str(e):
                # https://github.com/afawcett/apex-toolingapi/issues/19
                error += "Please verify if user has 'View All Data' permission."

            self.report.warning(message=error, exc=e)
        else:
            customFields = {
                record["DeveloperName"]: record
                for record in sObject_custom_fields_response["records"]
            }

        fields: List[SchemaFieldClass] = []
        primaryKeys: List[str] = []
        foreignKeys: List[ForeignKeyConstraintClass] = []

        for field in sObject_fields_response["records"]:
            customField = customFields.get(field["DeveloperName"], {})

            fieldName = field["QualifiedApiName"]
            fieldType = field["DataType"]

            # Skip compound fields. All Leaf fields are ingested instead.
            if fieldType in ("address", "location"):
                continue

            schemaField: SchemaFieldClass = self._get_schema_field(
                sObjectName, fieldName, fieldType, field, customField
            )
            fields.append(schemaField)

            if fieldType == "id":
                primaryKeys.append(fieldName)

            if (
                fieldType == "reference"
                and field["ReferenceTo"]["referenceTo"] is not None
            ):
                foreignKeys.extend(
                    list(self.get_foreign_keys_from_field(fieldName, field, datasetUrn))
                )

        schemaMetadata = SchemaMetadataClass(
            schemaName="",
            platform=builder.make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=OtherSchemaClass(rawSchema=""),
            fields=fields,
            primaryKeys=primaryKeys,
            foreignKeys=foreignKeys or None,
        )

        # Created Date and Actor are available for Custom Object only
        if customObject.get("CreatedDate") and customObject.get("CreatedBy"):
            schemaMetadata.created = self.get_audit_stamp(
                customObject["CreatedDate"], customObject["CreatedBy"]["Username"]
            )
        self.fieldCounts[sObjectName] = len(fields)

        yield MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=schemaMetadata
        ).as_workunit()

    def get_foreign_keys_from_field(
        self, fieldName: str, field: dict, datasetUrn: str
    ) -> Iterable[ForeignKeyConstraintClass]:
        # https://developer.salesforce.com/docs/atlas.en-us.object_reference.meta/object_reference/field_types.htm#i1435823
        foreignDatasets = [
            builder.make_dataset_urn_with_platform_instance(
                self.platform,
                fsObject,
                self.config.platform_instance,
                self.config.env,
            )
            for fsObject in field["ReferenceTo"]["referenceTo"]
        ]

        for foreignDataset in foreignDatasets:
            yield ForeignKeyConstraintClass(
                name=field["RelationshipName"] if field.get("RelationshipName") else "",
                foreignDataset=foreignDataset,
                foreignFields=[builder.make_schema_field_urn(foreignDataset, "Id")],
                sourceFields=[builder.make_schema_field_urn(datasetUrn, fieldName)],
            )

    def get_report(self) -> SourceReport:
        return self.report


def get_tags(params: Optional[List[str]] = None) -> GlobalTagsClass:
    if params is None:
        params = []
    tags = [TagAssociationClass(tag=builder.make_tag_urn(tag)) for tag in params if tag]
    return GlobalTagsClass(tags=tags)
